@@ -55,19 +55,47 @@ G_DEFINE_TYPE (GkdSecretUnlock, gkd_secret_unlock, GKD_SECRET_TYPE_PROMPT);
  * INTERNAL
  */
 
-static gchar*
-location_string_for_collection (GP11Object *collection)
+static GP11Attributes*
+attributes_for_collection (GP11Object *collection)
 {
-	gpointer identifier;
-	gsize n_identifier;
-	gchar *location;
+	GP11Attributes *attrs;
+	GError *error = NULL;
 
-	/* Figure out the identifier */
-	identifier = gp11_object_get_data (collection, CKA_ID, &n_identifier, NULL);
-	if (!identifier || !g_utf8_validate (identifier, n_identifier, NULL)) {
+	attrs = gp11_object_get (collection, &error, CKA_LABEL, CKA_ID, GP11_INVALID);
+	if (attrs == NULL) {
+		g_warning ("couldn't get attributes for collection: %s", error->message);
+		g_clear_error (&error);
+		return NULL;
+	}
+
+	return attrs;
+}
+
+static gchar*
+identifier_string_for_attributes (GP11Attributes *attrs)
+{
+	gchar *identifier;
+
+	g_assert (attrs);
+
+	if (!gp11_attributes_find_string (attrs, CKA_ID, &identifier))
+		return NULL;
+	if (!g_utf8_validate (identifier, -1, NULL)) {
 		g_free (identifier);
 		return NULL;
 	}
+	return identifier;
+}
+
+static gchar*
+location_string_for_attributes (GP11Attributes *attrs)
+{
+	gchar *identifier;
+	gchar *location;
+
+	identifier = identifier_string_for_attributes (attrs);
+	if (identifier == NULL)
+		return NULL;
 
 	/*
 	 * COMPAT: Format it into a string. This is done this way for compatibility
@@ -80,13 +108,53 @@ location_string_for_collection (GP11Object *collection)
 	return location;
 }
 
-static void
-prepare_unlock_prompt (GkdSecretUnlock *self, GP11Object *coll)
+static gchar*
+label_string_for_attributes (GP11Attributes *attrs)
 {
+	gchar *label;
+
+	if (!gp11_attributes_find_string (attrs, CKA_LABEL, &label))
+		label = NULL;
+	if (!label)
+		return g_strdup (_("Unnamed"));
+	else
+		return label;
+}
+
+static void
+prepare_unlock_login (GkdSecretUnlock *self)
+{
+	GkdPrompt *prompt;
+	const gchar *text;
+
+	g_assert (GKD_SECRET_IS_UNLOCK (self));
+
+	prompt = GKD_PROMPT (self);
+
+	gkd_prompt_set_title (prompt, _("Unlock Login Keyring"));
+
+	text = _("Enter password for to unlock your login keyring");
+	gkd_prompt_set_primary_text (prompt, text);
+
+	if (gkd_login_did_unlock_fail ())
+		text = _("The password you use to log in to your computer no longer matches that of your login keyring.");
+	else
+		text = _("The login keyring did not get unlocked when you logged into your computer.");
+	gkd_prompt_set_secondary_text (prompt, text);
+
+	gkd_prompt_hide_widget (prompt, "name_area");
+	gkd_prompt_hide_widget (prompt, "confirm_area");
+	gkd_prompt_show_widget (prompt, "password_area");
+}
+
+static void
+prepare_unlock_prompt (GkdSecretUnlock *self, GP11Object *coll, gboolean first)
+{
+	GP11Attributes *template;
+	GP11Attributes *attrs;
 	GError *error = NULL;
 	GkdPrompt *prompt;
-	gpointer data;
-	gsize n_data;
+	gchar *identifier;
 	gchar *label;
 	gchar *text;
 
@@ -95,19 +163,22 @@ prepare_unlock_prompt (GkdSecretUnlock *self, GP11Object *coll)
 
 	prompt = GKD_PROMPT (self);
 
-	data = gp11_object_get_data (coll, CKA_LABEL, &n_data, &error);
-	if (!data) {
-		g_warning ("couldn't get label for collection: %s", error->message);
-		g_clear_error (&error);
+	/* Hard reset on first prompt, soft on later */
+	gkd_prompt_reset (GKD_PROMPT (prompt), first);
+
+	attrs = attributes_for_collection (coll);
+	g_return_if_fail (attrs);
+
+	/* Login keyring is handled specially */
+	identifier = identifier_string_for_attributes (attrs);
+	if (identifier && g_str_equal (identifier, "login")) {
+		prepare_unlock_login (self);
+		g_free (identifier);
+		return;
 	}
 
-	if (!data || !n_data)
-		label = g_strdup (_("Unnamed"));
-	else
-		label = g_strndup (data, n_data);
-	g_free (data);
-
-	gkd_prompt_reset (prompt);
+	g_free (identifier);
+	label = label_string_for_attributes (attrs);
 
 	gkd_prompt_set_title (prompt, _("Unlock Keyring"));
 
@@ -121,10 +192,24 @@ prepare_unlock_prompt (GkdSecretUnlock *self, GP11Object *coll)
 
 	gkd_prompt_hide_widget (prompt, "name_area");
 	gkd_prompt_hide_widget (prompt, "confirm_area");
-	gkd_prompt_hide_widget (prompt, "details_area");
+	gkd_prompt_show_widget (prompt, "details_area");
 	gkd_prompt_show_widget (prompt, "password_area");
+	gkd_prompt_show_widget (prompt, "lock_area");
+	gkd_prompt_show_widget (prompt, "options_area");
 
 	g_free (label);
+
+	/* Setup the unlock options */
+	if (first) {
+		template = gp11_object_get_template (coll, CKA_G_CREDENTIAL_TEMPLATE, &error);
+		if (template) {
+			gkd_prompt_set_unlock_options (prompt, template);
+			gp11_attributes_unref (template);
+		} else {
+			g_warning ("couldn't get credential template for collection: %s", error->message);
+			g_clear_error (&error);
+		}
+	}
 }
 
 static void
@@ -154,21 +239,72 @@ check_locked_collection (GP11Object *collection, gboolean *locked)
 	return TRUE;
 }
 
+static void
+attach_credential_to_login (GP11Object *collection, GP11Object *cred)
+{
+	GError *error = NULL;
+	GP11Attributes *attrs;
+	gpointer value;
+	gsize n_value;
+	gchar *location;
+	gchar *label;
+
+	g_assert (GP11_IS_OBJECT (collection));
+	g_assert (GP11_IS_OBJECT (cred));
+
+	attrs = attributes_for_collection (collection);
+	g_return_if_fail (attrs);
+
+	location = location_string_for_attributes (attrs);
+	label = label_string_for_attributes (attrs);
+	gp11_attributes_unref (attrs);
+
+	value = gp11_object_get_data_full (cred, CKA_VALUE, egg_secure_realloc, NULL, &n_value, &error);
+	if (value) {
+		if (g_utf8_validate (value, n_value, NULL))
+			gkd_login_attach_secret (label, value, "keyring", location, NULL);
+		else
+			g_warning ("couldn't save non utf-8 unlock credentials in login keyring");
+		egg_secure_clear (value, n_value);
+		egg_secure_free (value);
+
+	} else {
+		g_warning ("couldn't read unlock credentials to save in login keyring: %s", error->message);
+		g_clear_error (&error);
+	}
+
+	g_free (location);
+	g_free (label);
+}
+
+static void
+common_unlock_attributes (GP11Attributes *attrs, GP11Object *collection)
+{
+	g_assert (attrs);
+	g_assert (GP11_IS_OBJECT (collection));
+	gp11_attributes_add_ulong (attrs, CKA_CLASS, CKO_G_CREDENTIAL);
+	gp11_attributes_add_ulong (attrs, CKA_G_OBJECT, gp11_object_get_handle (collection));
+}
+
 static gboolean
 authenticate_collection (GkdSecretUnlock *self, GP11Object *collection, gboolean *locked)
 {
 	DBusError derr = DBUS_ERROR_INIT;
 	GkdSecretSecret *master;
-	gboolean result;
+	GP11Attributes *template;
+	GP11Object *cred;
+	gboolean transient;
 
 	g_assert (GKD_SECRET_IS_UNLOCK (self));
 	g_assert (GP11_IS_OBJECT (collection));
 	g_assert (locked);
 
-	/* Bail out early, just checking locked status */
-	if (!gkd_prompt_has_response (GKD_PROMPT (self))) {
-		return check_locked_collection (collection, locked);
-	}
+	if (!check_locked_collection (collection, locked))
+		return FALSE;
+
+	/* Shortcut if already unlocked, or just checking locked status */
+	if (!*locked || !gkd_prompt_has_response (GKD_PROMPT (self)))
+		return TRUE;
 
 	master = gkd_secret_prompt_get_secret (GKD_SECRET_PROMPT (self), "password");
 	if (master == NULL) {
@@ -176,14 +312,33 @@ authenticate_collection (GkdSecretUnlock *self, GP11Object *collection, gboolean
 		return FALSE;
 	}
 
-	result = gkd_secret_unlock_with_secret (collection, master, &derr);
+	/* The various unlock options */
+	template = gp11_attributes_new ();
+	common_unlock_attributes (template, collection);
+	gkd_prompt_get_unlock_options (GKD_PROMPT (self), template);
+
+	/* If it's supposed to save non-transient, then we override that */
+	if (!gp11_attributes_find_boolean (template, CKA_GNOME_TRANSIENT, &transient))
+		transient = TRUE;
+
+	cred = gkd_secret_session_create_credential (master->session, NULL, template, master, &derr);
 	gkd_secret_secret_free (master);
 
-	if (result) {
+	if (cred) {
+		/* Save it to the login keyring */
+		if (!transient)
+			attach_credential_to_login (collection, cred);
+		g_object_unref (cred);
+
+		/* Save away the unlock options for next time */
+		gp11_object_set_template (collection, CKA_G_CREDENTIAL_TEMPLATE, template, NULL);
+		gp11_attributes_unref (template);
+
 		*locked = FALSE;
 		return TRUE; /* Operation succeeded, and unlocked */
 
 	} else {
+		gp11_attributes_unref (template);
 		if (dbus_error_has_name (&derr, INTERNAL_ERROR_DENIED)) {
 			dbus_error_free (&derr);
 			*locked = TRUE;
@@ -227,7 +382,7 @@ gkd_secret_unlock_prompt_ready (GkdSecretPrompt *prompt)
 
 			/* Collection still locked, prompt again */
 			} else if (locked) {
-				prepare_unlock_prompt (self, coll);
+				prepare_unlock_prompt (self, coll, FALSE);
 				set_warning_wrong (self);
 
 			/* Collection not locked, done with this one */
@@ -268,7 +423,7 @@ gkd_secret_unlock_prompt_ready (GkdSecretPrompt *prompt)
 			continue;
 		}
 
-		prepare_unlock_prompt (self, coll);
+		prepare_unlock_prompt (self, coll, TRUE);
 		g_object_unref (coll);
 		self->current = objpath;
 	}
@@ -351,6 +506,7 @@ void
 gkd_secret_unlock_queue (GkdSecretUnlock *self, const gchar *objpath)
 {
 	gboolean locked = TRUE;
+	GP11Attributes *attrs;
 	GP11Object *coll;
 	gchar *password;
 	gchar *location;
@@ -369,7 +525,10 @@ gkd_secret_unlock_queue (GkdSecretUnlock *self, const gchar *objpath)
 
 	/* Or try to use login keyring's passwords */
 	} else {
-		location = location_string_for_collection (coll);
+		attrs = attributes_for_collection (coll);
+		location = location_string_for_attributes (attrs);
+		gp11_attributes_unref (attrs);
+
 		if (location) {
 			password = gkd_login_lookup_secret ("keyring", location, NULL);
 			g_free (location);
@@ -434,11 +593,10 @@ gkd_secret_unlock_with_secret (GP11Object *collection, GkdSecretSecret *master,
 	if (check_locked_collection (collection, &locked) && !locked)
 		return TRUE;
 
-	attrs = gp11_attributes_newv (CKA_CLASS, GP11_ULONG, CKO_G_CREDENTIAL,
-	                              CKA_G_OBJECT, GP11_ULONG, gp11_object_get_handle (collection),
-	                              CKA_GNOME_TRANSIENT, GP11_BOOLEAN, TRUE,
-	                              CKA_TOKEN, GP11_BOOLEAN, TRUE,
-	                              GP11_INVALID);
+	attrs = gp11_attributes_new ();
+	common_unlock_attributes (attrs, collection);
+	gp11_attributes_add_boolean (attrs, CKA_GNOME_TRANSIENT, TRUE);
+	gp11_attributes_add_boolean (attrs, CKA_TOKEN, TRUE);
 
 	cred = gkd_secret_session_create_credential (master->session, NULL, attrs, master, derr);
 
@@ -453,6 +611,7 @@ gboolean
 gkd_secret_unlock_with_password (GP11Object *collection, const guchar *password,
                                  gsize n_password, DBusError *derr)
 {
+	GP11Attributes *attrs;
 	GError *error = NULL;
 	GP11Session *session;
 	GP11Object *cred;
@@ -467,13 +626,13 @@ gkd_secret_unlock_with_password (GP11Object *collection, const guchar *password,
 	session = gp11_object_get_session (collection);
 	g_return_val_if_fail (session, FALSE);
 
-	cred = gp11_session_create_object (session, &error, CKA_CLASS, GP11_ULONG, CKO_G_CREDENTIAL,
-	                                   CKA_G_OBJECT, GP11_ULONG, gp11_object_get_handle (collection),
-	                                   CKA_GNOME_TRANSIENT, GP11_BOOLEAN, TRUE,
-	                                   CKA_TOKEN, GP11_BOOLEAN, TRUE,
-	                                   CKA_VALUE, n_password, password,
-	                                   GP11_INVALID);
+	attrs = gp11_attributes_new_full (egg_secure_realloc);
+	common_unlock_attributes (attrs, collection);
+	gp11_attributes_add_boolean (attrs, CKA_GNOME_TRANSIENT, TRUE);
+	gp11_attributes_add_boolean (attrs, CKA_TOKEN, TRUE);
+	gp11_attributes_add_data (attrs, CKA_VALUE, password, n_password);
 
+	cred = gp11_session_create_object_full (session, attrs, NULL, &error);
 	if (cred == NULL) {
 		if (error->code == CKR_PIN_INCORRECT) {
 			dbus_set_error_const (derr, INTERNAL_ERROR_DENIED, "The password was incorrect.");
