@@ -375,6 +375,7 @@ gp11_mechanism_info_free (GP11MechanismInfo *mech_info)
 /**
  * gp11_mechanisms_check:
  * @mechanisms: A list of mechanisms, perhaps retrieved from gp11_slot_get_mechanisms().
+ * @...: A list of mechanism types followed by GP11_INVALID.
  * 
  * Check whether all the mechanism types are in the list.
  * 
@@ -620,6 +621,7 @@ gp11_slot_get_token_info (GP11Slot *self)
 			tokeninfo->utc_time = -1;
 		else
 			tokeninfo->utc_time = timegm (&tm);
+		g_free (string);
 	} else {
 		tokeninfo->utc_time = -1;
 	}
@@ -822,6 +824,8 @@ typedef struct OpenSession {
 	GP11Arguments base;
 	GP11Slot *slot;
 	gulong flags;
+	gpointer app_data;
+	CK_NOTIFY notify;
 	gchar *password;
 	gboolean auto_login;
 	CK_SESSION_HANDLE session;
@@ -839,7 +843,7 @@ perform_open_session (OpenSession *args)
 	/* First step, open session */
 	if (!args->session) {
 		rv = (args->base.pkcs11->C_OpenSession) (args->base.handle, args->flags, 
-		                                         NULL, NULL, &args->session);
+		                                         args->app_data, args->notify, &args->session);
 	}
 
 	if (rv != CKR_OK || !args->auto_login)
@@ -910,13 +914,15 @@ free_open_session (OpenSession *args)
 GP11Session*
 gp11_slot_open_session (GP11Slot *self, gulong flags, GError **err)
 {
-	return gp11_slot_open_session_full (self, flags, NULL, err);
+	return gp11_slot_open_session_full (self, flags, NULL, NULL, NULL, err);
 }
 
 /**
  * gp11_slot_open_session_full:
  * @self: The slot to open a session on.
  * @flags: The flags to open a session with.
+ * @app_data: Application data for notification callback.
+ * @notify: PKCS#11 notification callback.
  * @cancellable: Optional cancellation object, or NULL.
  * @err: A location to return an error, or NULL.
  * 
@@ -928,7 +934,8 @@ gp11_slot_open_session (GP11Slot *self, gulong flags, GError **err)
  * Return value: A new session or NULL if an error occurs.
  **/
 GP11Session*
-gp11_slot_open_session_full (GP11Slot *self, gulong flags, GCancellable *cancellable, GError **err)
+gp11_slot_open_session_full (GP11Slot *self, gulong flags, gpointer app_data,
+                             CK_NOTIFY notify, GCancellable *cancellable, GError **err)
 {
 	GP11Session *session = NULL;
 	GP11Module *module = NULL;
@@ -952,6 +959,8 @@ gp11_slot_open_session_full (GP11Slot *self, gulong flags, GCancellable *cancell
 		
 		args.slot = self;
 		args.flags = flags;
+		args.app_data = app_data;
+		args.notify = notify;
 		args.password = NULL;
 		args.auto_login = (gp11_module_get_auto_authenticate (module) & GP11_AUTHENTICATE_TOKENS) ? TRUE : FALSE;
 		args.session = 0;
@@ -970,6 +979,8 @@ gp11_slot_open_session_full (GP11Slot *self, gulong flags, GCancellable *cancell
  * gp11_slot_open_session_async:
  * @self: The slot to open a session on.
  * @flags: The flags to open a session with.
+ * @app_data: Application data for notification callback.
+ * @notify: PKCS#11 notification callback.
  * @cancellable: Optional cancellation object, or NULL.
  * @callback: Called when the operation completes.
  * @user_data: Data to pass to the callback.
@@ -980,7 +991,8 @@ gp11_slot_open_session_full (GP11Slot *self, gulong flags, GCancellable *cancell
  * This call will return immediately and complete asynchronously.
  **/
 void
-gp11_slot_open_session_async (GP11Slot *self, gulong flags, GCancellable *cancellable, 
+gp11_slot_open_session_async (GP11Slot *self, gulong flags, gpointer app_data,
+                              CK_NOTIFY notify, GCancellable *cancellable,
                               GAsyncReadyCallback callback, gpointer user_data)
 {
 	GP11Module *module = NULL;
@@ -996,6 +1008,8 @@ gp11_slot_open_session_async (GP11Slot *self, gulong flags, GCancellable *cancel
 	                               sizeof (*args), free_open_session);
 
 	args->flags = flags;
+	args->app_data = app_data;
+	args->notify = notify;
 	args->slot = g_object_ref (self);
 
 	/* Try to use a cached session */
